# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Union, Optional
from datetime import datetime
from typing_extensions import Literal, Annotated, TypeAlias

from .price import Price
from .address import Address
from .invoice import Invoice
from .maximum import Maximum
from .minimum import Minimum
from ..._utils import PropertyInfo
from .discount import Discount
from ..._models import BaseModel
from .tax_amount import TaxAmount
from .credit_note import CreditNote
from .invoice_tiny import InvoiceTiny
from .customer_tax_id import CustomerTaxID
from .credit_note_tiny import CreditNoteTiny
from .customer_minified import CustomerMinified
from .tier_sub_line_item import TierSubLineItem
from .other_sub_line_item import OtherSubLineItem
from .matrix_sub_line_item import MatrixSubLineItem
from .subscription_minified import SubscriptionMinified
from .invoice_level_discount import InvoiceLevelDiscount
from .monetary_maximum_adjustment import MonetaryMaximumAdjustment
from .monetary_minimum_adjustment import MonetaryMinimumAdjustment
from .monetary_usage_discount_adjustment import MonetaryUsageDiscountAdjustment
from .monetary_amount_discount_adjustment import MonetaryAmountDiscountAdjustment
from .monetary_percentage_discount_adjustment import MonetaryPercentageDiscountAdjustment

__all__ = [
    "ChangedSubscriptionResources",
    "CreatedInvoice",
    "CreatedInvoiceAutoCollection",
    "CreatedInvoiceCreditNote",
    "CreatedInvoiceCustomerBalanceTransaction",
    "CreatedInvoiceLineItem",
    "CreatedInvoiceLineItemAdjustment",
    "CreatedInvoiceLineItemSubLineItem",
    "CreatedInvoicePaymentAttempt",
]


class CreatedInvoiceAutoCollection(BaseModel):
    enabled: Optional[bool] = None
    """True only if auto-collection is enabled for this invoice."""

    next_attempt_at: Optional[datetime] = None
    """
    If the invoice is scheduled for auto-collection, this field will reflect when
    the next attempt will occur. If dunning has been exhausted, or auto-collection
    is not enabled for this invoice, this field will be `null`.
    """

    num_attempts: Optional[int] = None
    """Number of auto-collection payment attempts."""

    previously_attempted_at: Optional[datetime] = None
    """
    If Orb has ever attempted payment auto-collection for this invoice, this field
    will reflect when that attempt occurred. In conjunction with `next_attempt_at`,
    this can be used to tell whether the invoice is currently in dunning (that is,
    `previously_attempted_at` is non-null, and `next_attempt_time` is non-null), or
    if dunning has been exhausted (`previously_attempted_at` is non-null, but
    `next_attempt_time` is null).
    """


class CreatedInvoiceCreditNote(BaseModel):
    id: str

    credit_note_number: str

    memo: Optional[str] = None
    """An optional memo supplied on the credit note."""

    reason: str

    total: str

    type: str

    voided_at: Optional[datetime] = None
    """
    If the credit note has a status of `void`, this gives a timestamp when the
    credit note was voided.
    """


class CreatedInvoiceCustomerBalanceTransaction(BaseModel):
    id: str
    """A unique id for this transaction."""

    action: Literal[
        "applied_to_invoice",
        "manual_adjustment",
        "prorated_refund",
        "revert_prorated_refund",
        "return_from_voiding",
        "credit_note_applied",
        "credit_note_voided",
        "overpayment_refund",
        "external_payment",
        "small_invoice_carryover",
    ]

    amount: str
    """The value of the amount changed in the transaction."""

    created_at: datetime
    """The creation time of this transaction."""

    credit_note: Optional[CreditNoteTiny] = None

    description: Optional[str] = None
    """An optional description provided for manual customer balance adjustments."""

    ending_balance: str
    """
    The new value of the customer's balance prior to the transaction, in the
    customer's currency.
    """

    invoice: Optional[InvoiceTiny] = None

    starting_balance: str
    """
    The original value of the customer's balance prior to the transaction, in the
    customer's currency.
    """

    type: Literal["increment", "decrement"]


CreatedInvoiceLineItemAdjustment: TypeAlias = Annotated[
    Union[
        MonetaryUsageDiscountAdjustment,
        MonetaryAmountDiscountAdjustment,
        MonetaryPercentageDiscountAdjustment,
        MonetaryMinimumAdjustment,
        MonetaryMaximumAdjustment,
    ],
    PropertyInfo(discriminator="adjustment_type"),
]

CreatedInvoiceLineItemSubLineItem: TypeAlias = Annotated[
    Union[MatrixSubLineItem, TierSubLineItem, OtherSubLineItem], PropertyInfo(discriminator="type")
]


class CreatedInvoiceLineItem(BaseModel):
    id: str
    """A unique ID for this line item."""

    adjusted_subtotal: str
    """
    The line amount after any adjustments and before overage conversion, credits and
    partial invoicing.
    """

    adjustments: List[CreatedInvoiceLineItemAdjustment]
    """
    All adjustments applied to the line item in the order they were applied based on
    invoice calculations (ie. usage discounts -> amount discounts -> percentage
    discounts -> minimums -> maximums).
    """

    amount: str
    """
    The final amount for a line item after all adjustments and pre paid credits have
    been applied.
    """

    credits_applied: str
    """The number of prepaid credits applied."""

    discount: Optional[Discount] = None
    """This field is deprecated in favor of `adjustments`"""

    end_date: datetime
    """The end date of the range of time applied for this line item's price."""

    filter: Optional[str] = None
    """An additional filter that was used to calculate the usage for this line item."""

    grouping: Optional[str] = None
    """
    [DEPRECATED] For configured prices that are split by a grouping key, this will
    be populated with the key and a value. The `amount` and `subtotal` will be the
    values for this particular grouping.
    """

    maximum: Optional[Maximum] = None
    """This field is deprecated in favor of `adjustments`."""

    maximum_amount: Optional[str] = None
    """This field is deprecated in favor of `adjustments`."""

    minimum: Optional[Minimum] = None
    """This field is deprecated in favor of `adjustments`."""

    minimum_amount: Optional[str] = None
    """This field is deprecated in favor of `adjustments`."""

    name: str
    """The name of the price associated with this line item."""

    partially_invoiced_amount: str
    """Any amount applied from a partial invoice"""

    price: Price
    """
    The Price resource represents a price that can be billed on a subscription,
    resulting in a charge on an invoice in the form of an invoice line item. Prices
    take a quantity and determine an amount to bill.

    Orb supports a few different pricing models out of the box. Each of these models
    is serialized differently in a given Price object. The model_type field
    determines the key for the configuration object that is present.

    For more on the types of prices, see
    [the core concepts documentation](/core-concepts#plan-and-price)
    """

    quantity: float
    """Either the fixed fee quantity or the usage during the service period."""

    start_date: datetime
    """The start date of the range of time applied for this line item's price."""

    sub_line_items: List[CreatedInvoiceLineItemSubLineItem]
    """
    For complex pricing structures, the line item can be broken down further in
    `sub_line_items`.
    """

    subtotal: str
    """The line amount before any adjustments."""

    tax_amounts: List[TaxAmount]
    """An array of tax rates and their incurred tax amounts.

    Empty if no tax integration is configured.
    """

    usage_customer_ids: Optional[List[str]] = None
    """
    A list of customer ids that were used to calculate the usage for this line item.
    """


class CreatedInvoicePaymentAttempt(BaseModel):
    id: str
    """The ID of the payment attempt."""

    amount: str
    """The amount of the payment attempt."""

    created_at: datetime
    """The time at which the payment attempt was created."""

    payment_provider: Optional[Literal["stripe"]] = None
    """The payment provider that attempted to collect the payment."""

    payment_provider_id: Optional[str] = None
    """The ID of the payment attempt in the payment provider."""

    receipt_pdf: Optional[str] = None
    """URL to the downloadable PDF version of the receipt.

    This field will be `null` for payment attempts that did not succeed.
    """

    succeeded: bool
    """Whether the payment attempt succeeded."""


class CreatedInvoice(BaseModel):
    id: str

    amount_due: str
    """
    This is the final amount required to be charged to the customer and reflects the
    application of the customer balance to the `total` of the invoice.
    """

    auto_collection: CreatedInvoiceAutoCollection

    billing_address: Optional[Address] = None

    created_at: datetime
    """The creation time of the resource in Orb."""

    credit_notes: List[CreatedInvoiceCreditNote]
    """A list of credit notes associated with the invoice"""

    currency: str
    """An ISO 4217 currency string or `credits`"""

    customer: CustomerMinified

    customer_balance_transactions: List[CreatedInvoiceCustomerBalanceTransaction]

    customer_tax_id: Optional[CustomerTaxID] = None
    """
    Tax IDs are commonly required to be displayed on customer invoices, which are
    added to the headers of invoices.

    ### Supported Tax ID Countries and Types

    | Country                | Type         | Description                                                                                             |
    | ---------------------- | ------------ | ------------------------------------------------------------------------------------------------------- |
    | Albania                | `al_tin`     | Albania Tax Identification Number                                                                       |
    | Andorra                | `ad_nrt`     | Andorran NRT Number                                                                                     |
    | Angola                 | `ao_tin`     | Angola Tax Identification Number                                                                        |
    | Argentina              | `ar_cuit`    | Argentinian Tax ID Number                                                                               |
    | Armenia                | `am_tin`     | Armenia Tax Identification Number                                                                       |
    | Aruba                  | `aw_tin`     | Aruba Tax Identification Number                                                                         |
    | Australia              | `au_abn`     | Australian Business Number (AU ABN)                                                                     |
    | Australia              | `au_arn`     | Australian Taxation Office Reference Number                                                             |
    | Austria                | `eu_vat`     | European VAT Number                                                                                     |
    | Azerbaijan             | `az_tin`     | Azerbaijan Tax Identification Number                                                                    |
    | Bahamas                | `bs_tin`     | Bahamas Tax Identification Number                                                                       |
    | Bahrain                | `bh_vat`     | Bahraini VAT Number                                                                                     |
    | Bangladesh             | `bd_bin`     | Bangladesh Business Identification Number                                                               |
    | Barbados               | `bb_tin`     | Barbados Tax Identification Number                                                                      |
    | Belarus                | `by_tin`     | Belarus TIN Number                                                                                      |
    | Belgium                | `eu_vat`     | European VAT Number                                                                                     |
    | Benin                  | `bj_ifu`     | Benin Tax Identification Number (Identifiant Fiscal Unique)                                             |
    | Bolivia                | `bo_tin`     | Bolivian Tax ID                                                                                         |
    | Bosnia and Herzegovina | `ba_tin`     | Bosnia and Herzegovina Tax Identification Number                                                        |
    | Brazil                 | `br_cnpj`    | Brazilian CNPJ Number                                                                                   |
    | Brazil                 | `br_cpf`     | Brazilian CPF Number                                                                                    |
    | Bulgaria               | `bg_uic`     | Bulgaria Unified Identification Code                                                                    |
    | Bulgaria               | `eu_vat`     | European VAT Number                                                                                     |
    | Burkina Faso           | `bf_ifu`     | Burkina Faso Tax Identification Number (Numéro d'Identifiant Fiscal Unique)                             |
    | Cambodia               | `kh_tin`     | Cambodia Tax Identification Number                                                                      |
    | Cameroon               | `cm_niu`     | Cameroon Tax Identification Number (Numéro d'Identifiant fiscal Unique)                                 |
    | Canada                 | `ca_bn`      | Canadian BN                                                                                             |
    | Canada                 | `ca_gst_hst` | Canadian GST/HST Number                                                                                 |
    | Canada                 | `ca_pst_bc`  | Canadian PST Number (British Columbia)                                                                  |
    | Canada                 | `ca_pst_mb`  | Canadian PST Number (Manitoba)                                                                          |
    | Canada                 | `ca_pst_sk`  | Canadian PST Number (Saskatchewan)                                                                      |
    | Canada                 | `ca_qst`     | Canadian QST Number (Québec)                                                                            |
    | Cape Verde             | `cv_nif`     | Cape Verde Tax Identification Number (Número de Identificação Fiscal)                                   |
    | Chile                  | `cl_tin`     | Chilean TIN                                                                                             |
    | China                  | `cn_tin`     | Chinese Tax ID                                                                                          |
    | Colombia               | `co_nit`     | Colombian NIT Number                                                                                    |
    | Congo-Kinshasa         | `cd_nif`     | Congo (DR) Tax Identification Number (Número de Identificação Fiscal)                                   |
    | Costa Rica             | `cr_tin`     | Costa Rican Tax ID                                                                                      |
    | Croatia                | `eu_vat`     | European VAT Number                                                                                     |
    | Croatia                | `hr_oib`     | Croatian Personal Identification Number (OIB)                                                           |
    | Cyprus                 | `eu_vat`     | European VAT Number                                                                                     |
    | Czech Republic         | `eu_vat`     | European VAT Number                                                                                     |
    | Denmark                | `eu_vat`     | European VAT Number                                                                                     |
    | Dominican Republic     | `do_rcn`     | Dominican RCN Number                                                                                    |
    | Ecuador                | `ec_ruc`     | Ecuadorian RUC Number                                                                                   |
    | Egypt                  | `eg_tin`     | Egyptian Tax Identification Number                                                                      |
    | El Salvador            | `sv_nit`     | El Salvadorian NIT Number                                                                               |
    | Estonia                | `eu_vat`     | European VAT Number                                                                                     |
    | Ethiopia               | `et_tin`     | Ethiopia Tax Identification Number                                                                      |
    | European Union         | `eu_oss_vat` | European One Stop Shop VAT Number for non-Union scheme                                                  |
    | Finland                | `eu_vat`     | European VAT Number                                                                                     |
    | France                 | `eu_vat`     | European VAT Number                                                                                     |
    | Georgia                | `ge_vat`     | Georgian VAT                                                                                            |
    | Germany                | `de_stn`     | German Tax Number (Steuernummer)                                                                        |
    | Germany                | `eu_vat`     | European VAT Number                                                                                     |
    | Greece                 | `eu_vat`     | European VAT Number                                                                                     |
    | Guinea                 | `gn_nif`     | Guinea Tax Identification Number (Número de Identificação Fiscal)                                       |
    | Hong Kong              | `hk_br`      | Hong Kong BR Number                                                                                     |
    | Hungary                | `eu_vat`     | European VAT Number                                                                                     |
    | Hungary                | `hu_tin`     | Hungary Tax Number (adószám)                                                                            |
    | Iceland                | `is_vat`     | Icelandic VAT                                                                                           |
    | India                  | `in_gst`     | Indian GST Number                                                                                       |
    | Indonesia              | `id_npwp`    | Indonesian NPWP Number                                                                                  |
    | Ireland                | `eu_vat`     | European VAT Number                                                                                     |
    | Israel                 | `il_vat`     | Israel VAT                                                                                              |
    | Italy                  | `eu_vat`     | European VAT Number                                                                                     |
    | Japan                  | `jp_cn`      | Japanese Corporate Number (_Hōjin Bangō_)                                                               |
    | Japan                  | `jp_rn`      | Japanese Registered Foreign Businesses' Registration Number (_Tōroku Kokugai Jigyōsha no Tōroku Bangō_) |
    | Japan                  | `jp_trn`     | Japanese Tax Registration Number (_Tōroku Bangō_)                                                       |
    | Kazakhstan             | `kz_bin`     | Kazakhstani Business Identification Number                                                              |
    | Kenya                  | `ke_pin`     | Kenya Revenue Authority Personal Identification Number                                                  |
    | Kyrgyzstan             | `kg_tin`     | Kyrgyzstan Tax Identification Number                                                                    |
    | Laos                   | `la_tin`     | Laos Tax Identification Number                                                                          |
    | Latvia                 | `eu_vat`     | European VAT Number                                                                                     |
    | Liechtenstein          | `li_uid`     | Liechtensteinian UID Number                                                                             |
    | Liechtenstein          | `li_vat`     | Liechtenstein VAT Number                                                                                |
    | Lithuania              | `eu_vat`     | European VAT Number                                                                                     |
    | Luxembourg             | `eu_vat`     | European VAT Number                                                                                     |
    | Malaysia               | `my_frp`     | Malaysian FRP Number                                                                                    |
    | Malaysia               | `my_itn`     | Malaysian ITN                                                                                           |
    | Malaysia               | `my_sst`     | Malaysian SST Number                                                                                    |
    | Malta                  | `eu_vat`     | European VAT Number                                                                                     |
    | Mauritania             | `mr_nif`     | Mauritania Tax Identification Number (Número de Identificação Fiscal)                                   |
    | Mexico                 | `mx_rfc`     | Mexican RFC Number                                                                                      |
    | Moldova                | `md_vat`     | Moldova VAT Number                                                                                      |
    | Montenegro             | `me_pib`     | Montenegro PIB Number                                                                                   |
    | Morocco                | `ma_vat`     | Morocco VAT Number                                                                                      |
    | Nepal                  | `np_pan`     | Nepal PAN Number                                                                                        |
    | Netherlands            | `eu_vat`     | European VAT Number                                                                                     |
    | New Zealand            | `nz_gst`     | New Zealand GST Number                                                                                  |
    | Nigeria                | `ng_tin`     | Nigerian Tax Identification Number                                                                      |
    | North Macedonia        | `mk_vat`     | North Macedonia VAT Number                                                                              |
    | Northern Ireland       | `eu_vat`     | Northern Ireland VAT Number                                                                             |
    | Norway                 | `no_vat`     | Norwegian VAT Number                                                                                    |
    | Norway                 | `no_voec`    | Norwegian VAT on e-commerce Number                                                                      |
    | Oman                   | `om_vat`     | Omani VAT Number                                                                                        |
    | Peru                   | `pe_ruc`     | Peruvian RUC Number                                                                                     |
    | Philippines            | `ph_tin`     | Philippines Tax Identification Number                                                                   |
    | Poland                 | `eu_vat`     | European VAT Number                                                                                     |
    | Portugal               | `eu_vat`     | European VAT Number                                                                                     |
    | Romania                | `eu_vat`     | European VAT Number                                                                                     |
    | Romania                | `ro_tin`     | Romanian Tax ID Number                                                                                  |
    | Russia                 | `ru_inn`     | Russian INN                                                                                             |
    | Russia                 | `ru_kpp`     | Russian KPP                                                                                             |
    | Saudi Arabia           | `sa_vat`     | Saudi Arabia VAT                                                                                        |
    | Senegal                | `sn_ninea`   | Senegal NINEA Number                                                                                    |
    | Serbia                 | `rs_pib`     | Serbian PIB Number                                                                                      |
    | Singapore              | `sg_gst`     | Singaporean GST                                                                                         |
    | Singapore              | `sg_uen`     | Singaporean UEN                                                                                         |
    | Slovakia               | `eu_vat`     | European VAT Number                                                                                     |
    | Slovenia               | `eu_vat`     | European VAT Number                                                                                     |
    | Slovenia               | `si_tin`     | Slovenia Tax Number (davčna številka)                                                                   |
    | South Africa           | `za_vat`     | South African VAT Number                                                                                |
    | South Korea            | `kr_brn`     | Korean BRN                                                                                              |
    | Spain                  | `es_cif`     | Spanish NIF Number (previously Spanish CIF Number)                                                      |
    | Spain                  | `eu_vat`     | European VAT Number                                                                                     |
    | Suriname               | `sr_fin`     | Suriname FIN Number                                                                                     |
    | Sweden                 | `eu_vat`     | European VAT Number                                                                                     |
    | Switzerland            | `ch_uid`     | Switzerland UID Number                                                                                  |
    | Switzerland            | `ch_vat`     | Switzerland VAT Number                                                                                  |
    | Taiwan                 | `tw_vat`     | Taiwanese VAT                                                                                           |
    | Tajikistan             | `tj_tin`     | Tajikistan Tax Identification Number                                                                    |
    | Tanzania               | `tz_vat`     | Tanzania VAT Number                                                                                     |
    | Thailand               | `th_vat`     | Thai VAT                                                                                                |
    | Turkey                 | `tr_tin`     | Turkish Tax Identification Number                                                                       |
    | Uganda                 | `ug_tin`     | Uganda Tax Identification Number                                                                        |
    | Ukraine                | `ua_vat`     | Ukrainian VAT                                                                                           |
    | United Arab Emirates   | `ae_trn`     | United Arab Emirates TRN                                                                                |
    | United Kingdom         | `gb_vat`     | United Kingdom VAT Number                                                                               |
    | United States          | `us_ein`     | United States EIN                                                                                       |
    | Uruguay                | `uy_ruc`     | Uruguayan RUC Number                                                                                    |
    | Uzbekistan             | `uz_tin`     | Uzbekistan TIN Number                                                                                   |
    | Uzbekistan             | `uz_vat`     | Uzbekistan VAT Number                                                                                   |
    | Venezuela              | `ve_rif`     | Venezuelan RIF Number                                                                                   |
    | Vietnam                | `vn_tin`     | Vietnamese Tax ID Number                                                                                |
    | Zambia                 | `zm_tin`     | Zambia Tax Identification Number                                                                        |
    | Zimbabwe               | `zw_tin`     | Zimbabwe Tax Identification Number                                                                      |
    """

    discount: object
    """This field is deprecated in favor of `discounts`.

    If a `discounts` list is provided, the first discount in the list will be
    returned. If the list is empty, `None` will be returned.
    """

    discounts: List[InvoiceLevelDiscount]

    due_date: Optional[datetime] = None
    """When the invoice payment is due.

    The due date is null if the invoice is not yet finalized.
    """

    eligible_to_issue_at: Optional[datetime] = None
    """
    If the invoice has a status of `draft`, this will be the time that the invoice
    will be eligible to be issued, otherwise it will be `null`. If `auto-issue` is
    true, the invoice will automatically begin issuing at this time.
    """

    hosted_invoice_url: Optional[str] = None
    """A URL for the customer-facing invoice portal.

    This URL expires 30 days after the invoice's due date, or 60 days after being
    re-generated through the UI.
    """

    invoice_date: datetime
    """The scheduled date of the invoice"""

    invoice_number: str
    """Automatically generated invoice number to help track and reconcile invoices.

    Invoice numbers have a prefix such as `RFOBWG`. These can be sequential per
    account or customer.
    """

    invoice_pdf: Optional[str] = None
    """The link to download the PDF representation of the `Invoice`."""

    invoice_source: Literal["subscription", "partial", "one_off"]

    is_payable_now: bool
    """True if the invoice has only in-advance fixed fees and is payable now"""

    issue_failed_at: Optional[datetime] = None
    """
    If the invoice failed to issue, this will be the last time it failed to issue
    (even if it is now in a different state.)
    """

    issued_at: Optional[datetime] = None
    """
    If the invoice has been issued, this will be the time it transitioned to
    `issued` (even if it is now in a different state.)
    """

    line_items: List[CreatedInvoiceLineItem]
    """The breakdown of prices in this invoice."""

    maximum: Optional[Maximum] = None

    maximum_amount: Optional[str] = None

    memo: Optional[str] = None
    """
    Free-form text which is available on the invoice PDF and the Orb invoice portal.
    """

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    minimum: Optional[Minimum] = None

    minimum_amount: Optional[str] = None

    paid_at: Optional[datetime] = None
    """
    If the invoice has a status of `paid`, this gives a timestamp when the invoice
    was paid.
    """

    payment_attempts: List[CreatedInvoicePaymentAttempt]
    """A list of payment attempts associated with the invoice"""

    payment_failed_at: Optional[datetime] = None
    """
    If payment was attempted on this invoice but failed, this will be the time of
    the most recent attempt.
    """

    payment_started_at: Optional[datetime] = None
    """
    If payment was attempted on this invoice, this will be the start time of the
    most recent attempt. This field is especially useful for delayed-notification
    payment mechanisms (like bank transfers), where payment can take 3 days or more.
    """

    scheduled_issue_at: Optional[datetime] = None
    """
    If the invoice is in draft, this timestamp will reflect when the invoice is
    scheduled to be issued.
    """

    shipping_address: Optional[Address] = None

    status: Literal["issued", "paid", "synced", "void", "draft"]

    subscription: Optional[SubscriptionMinified] = None

    subtotal: str
    """The total before any discounts and minimums are applied."""

    sync_failed_at: Optional[datetime] = None
    """
    If the invoice failed to sync, this will be the last time an external invoicing
    provider sync was attempted. This field will always be `null` for invoices using
    Orb Invoicing.
    """

    total: str
    """The total after any minimums and discounts have been applied."""

    voided_at: Optional[datetime] = None
    """
    If the invoice has a status of `void`, this gives a timestamp when the invoice
    was voided.
    """

    will_auto_issue: bool
    """
    This is true if the invoice will be automatically issued in the future, and
    false otherwise.
    """


class ChangedSubscriptionResources(BaseModel):
    created_credit_notes: List[CreditNote]
    """The credit notes that were created as part of this operation."""

    created_invoices: List[CreatedInvoice]
    """The invoices that were created as part of this operation."""

    voided_credit_notes: List[CreditNote]
    """The credit notes that were voided as part of this operation."""

    voided_invoices: List[Invoice]
    """The invoices that were voided as part of this operation."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["BetaSetDefaultPlanVersionParams"]


class BetaSetDefaultPlanVersionParams(TypedDict, total=False):
    version: Required[int]
    """Plan version to set as the default."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .transform_price_filter import TransformPriceFilter

__all__ = ["PercentageDiscount"]


class PercentageDiscount(BaseModel):
    discount_type: Literal["percentage"]

    percentage_discount: float
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    applies_to_price_ids: Optional[List[str]] = None
    """List of price_ids that this discount applies to.

    For plan/plan phase discounts, this can be a subset of prices.
    """

    filters: Optional[List[TransformPriceFilter]] = None
    """The filters that determine which prices to apply this discount to."""

    reason: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime

from .price import Price
from ..._models import BaseModel
from .fixed_fee_quantity_transition import FixedFeeQuantityTransition

__all__ = ["PriceInterval"]


class PriceInterval(BaseModel):
    id: str

    billing_cycle_day: int
    """The day of the month that Orb bills for this price"""

    current_billing_period_end_date: Optional[datetime] = None
    """The end of the current billing period.

    This is an exclusive timestamp, such that the instant returned is exactly the
    end of the billing period. Set to null if this price interval is not currently
    active.
    """

    current_billing_period_start_date: Optional[datetime] = None
    """The start date of the current billing period.

    This is an inclusive timestamp; the instant returned is exactly the beginning of
    the billing period. Set to null if this price interval is not currently active.
    """

    end_date: Optional[datetime] = None
    """The end date of the price interval.

    This is the date that Orb stops billing for this price.
    """

    filter: Optional[str] = None
    """An additional filter to apply to usage queries."""

    fixed_fee_quantity_transitions: Optional[List[FixedFeeQuantityTransition]] = None
    """The fixed fee quantity transitions for this price interval.

    This is only relevant for fixed fees.
    """

    price: Price
    """
    The Price resource represents a price that can be billed on a subscription,
    resulting in a charge on an invoice in the form of an invoice line item. Prices
    take a quantity and determine an amount to bill.

    Orb supports a few different pricing models out of the box. Each of these models
    is serialized differently in a given Price object. The model_type field
    determines the key for the configuration object that is present.

    For more on the types of prices, see
    [the core concepts documentation](/core-concepts#plan-and-price)
    """

    start_date: datetime
    """The start date of the price interval.

    This is the date that Orb starts billing for this price.
    """

    usage_customer_ids: Optional[List[str]] = None
    """
    A list of customer IDs whose usage events will be aggregated and billed under
    this price interval.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypeAlias, TypedDict

from .._types import SequenceNotStr
from .._utils import PropertyInfo
from .discount_override_param import DiscountOverrideParam
from .shared_params.new_maximum import NewMaximum
from .shared_params.new_minimum import NewMinimum
from .shared_params.new_usage_discount import NewUsageDiscount
from .new_subscription_bulk_price_param import NewSubscriptionBulkPriceParam
from .new_subscription_unit_price_param import NewSubscriptionUnitPriceParam
from .shared_params.new_amount_discount import NewAmountDiscount
from .shared_params.new_allocation_price import NewAllocationPrice
from .new_subscription_matrix_price_param import NewSubscriptionMatrixPriceParam
from .new_subscription_tiered_price_param import NewSubscriptionTieredPriceParam
from .new_subscription_package_price_param import NewSubscriptionPackagePriceParam
from .shared_params.new_percentage_discount import NewPercentageDiscount
from .shared_params.unit_conversion_rate_config import UnitConversionRateConfig
from .new_subscription_grouped_tiered_price_param import NewSubscriptionGroupedTieredPriceParam
from .new_subscription_tiered_package_price_param import NewSubscriptionTieredPackagePriceParam
from .shared_params.tiered_conversion_rate_config import TieredConversionRateConfig
from .shared_params.new_billing_cycle_configuration import NewBillingCycleConfiguration
from .new_subscription_minimum_composite_price_param import NewSubscriptionMinimumCompositePriceParam
from .new_subscription_unit_with_percent_price_param import NewSubscriptionUnitWithPercentPriceParam
from .new_subscription_grouped_allocation_price_param import NewSubscriptionGroupedAllocationPriceParam
from .new_subscription_bulk_with_proration_price_param import NewSubscriptionBulkWithProrationPriceParam
from .new_subscription_tiered_with_minimum_price_param import NewSubscriptionTieredWithMinimumPriceParam
from .new_subscription_unit_with_proration_price_param import NewSubscriptionUnitWithProrationPriceParam
from .shared_params.billing_cycle_anchor_configuration import BillingCycleAnchorConfiguration
from .shared_params.new_dimensional_price_configuration import NewDimensionalPriceConfiguration
from .new_subscription_grouped_tiered_package_price_param import NewSubscriptionGroupedTieredPackagePriceParam
from .new_subscription_matrix_with_allocation_price_param import NewSubscriptionMatrixWithAllocationPriceParam
from .new_subscription_threshold_total_amount_price_param import NewSubscriptionThresholdTotalAmountPriceParam
from .new_subscription_cumulative_grouped_bulk_price_param import NewSubscriptionCumulativeGroupedBulkPriceParam
from .new_subscription_package_with_allocation_price_param import NewSubscriptionPackageWithAllocationPriceParam
from .new_subscription_matrix_with_display_name_price_param import NewSubscriptionMatrixWithDisplayNamePriceParam
from .new_subscription_max_group_tiered_package_price_param import NewSubscriptionMaxGroupTieredPackagePriceParam
from .new_subscription_tiered_package_with_minimum_price_param import NewSubscriptionTieredPackageWithMinimumPriceParam
from .new_subscription_grouped_with_metered_minimum_price_param import (
    NewSubscriptionGroupedWithMeteredMinimumPriceParam,
)
from .new_subscription_grouped_with_prorated_minimum_price_param import (
    NewSubscriptionGroupedWithProratedMinimumPriceParam,
)
from .new_subscription_scalable_matrix_with_unit_pricing_price_param import (
    NewSubscriptionScalableMatrixWithUnitPricingPriceParam,
)
from .new_subscription_scalable_matrix_with_tiered_pricing_price_param import (
    NewSubscriptionScalableMatrixWithTieredPricingPriceParam,
)

__all__ = [
    "SubscriptionCreateParams",
    "AddAdjustment",
    "AddAdjustmentAdjustment",
    "AddPrice",
    "AddPricePrice",
    "AddPricePriceNewSubscriptionTieredWithProrationPrice",
    "AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig",
    "AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier",
    "AddPricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig",
    "AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice",
    "AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig",
    "AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig",
    "RemoveAdjustment",
    "RemovePrice",
    "ReplaceAdjustment",
    "ReplaceAdjustmentAdjustment",
    "ReplacePrice",
    "ReplacePricePrice",
    "ReplacePricePriceNewSubscriptionTieredWithProrationPrice",
    "ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig",
    "ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier",
    "ReplacePricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig",
    "ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice",
    "ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig",
    "ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig",
]


class SubscriptionCreateParams(TypedDict, total=False):
    add_adjustments: Optional[Iterable[AddAdjustment]]
    """Additional adjustments to be added to the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    add_prices: Optional[Iterable[AddPrice]]
    """Additional prices to be added to the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    align_billing_with_subscription_start_date: bool

    auto_collection: Optional[bool]
    """
    Determines whether issued invoices for this subscription will automatically be
    charged with the saved payment method on the due date. If not specified, this
    defaults to the behavior configured for this customer.
    """

    aws_region: Optional[str]

    billing_cycle_anchor_configuration: Optional[BillingCycleAnchorConfiguration]

    coupon_redemption_code: Optional[str]
    """Redemption code to be used for this subscription.

    If the coupon cannot be found by its redemption code, or cannot be redeemed, an
    error response will be returned and the subscription creation or plan change
    will not be scheduled.
    """

    credits_overage_rate: Optional[float]

    currency: Optional[str]
    """The currency to use for the subscription.

    If not specified, the invoicing currency for the plan will be used.
    """

    customer_id: Optional[str]

    default_invoice_memo: Optional[str]
    """Determines the default memo on this subscription's invoices.

    Note that if this is not provided, it is determined by the plan configuration.
    """

    end_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    external_customer_id: Optional[str]

    external_marketplace: Optional[Literal["google", "aws", "azure"]]

    external_marketplace_reporting_id: Optional[str]

    external_plan_id: Optional[str]
    """
    The external_plan_id of the plan that the given subscription should be switched
    to. Note that either this property or `plan_id` must be specified.
    """

    filter: Optional[str]
    """An additional filter to apply to usage queries.

    This filter must be expressed as a boolean
    [computed property](/extensibility/advanced-metrics#computed-properties). If
    null, usage queries will not include any additional filter.
    """

    initial_phase_order: Optional[int]
    """The phase of the plan to start with"""

    invoicing_threshold: Optional[str]
    """
    When this subscription's accrued usage reaches this threshold, an invoice will
    be issued for the subscription. If not specified, invoices will only be issued
    at the end of the billing period.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    name: Optional[str]
    """The name to use for the subscription.

    If not specified, the plan name will be used.
    """

    net_terms: Optional[int]
    """
    The net terms determines the difference between the invoice date and the issue
    date for the invoice. If you intend the invoice to be due on issue, set this
    to 0. If not provided, this defaults to the value specified in the plan.
    """

    per_credit_overage_amount: Optional[float]

    plan_id: Optional[str]
    """The plan that the given subscription should be switched to.

    Note that either this property or `external_plan_id` must be specified.
    """

    plan_version_number: Optional[int]
    """Specifies which version of the plan to subscribe to.

    If null, the default version will be used.
    """

    price_overrides: Optional[Iterable[object]]
    """Optionally provide a list of overrides for prices on the plan"""

    remove_adjustments: Optional[Iterable[RemoveAdjustment]]
    """Plan adjustments to be removed from the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    remove_prices: Optional[Iterable[RemovePrice]]
    """Plan prices to be removed from the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    replace_adjustments: Optional[Iterable[ReplaceAdjustment]]
    """Plan adjustments to be replaced with additional adjustments on the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    replace_prices: Optional[Iterable[ReplacePrice]]
    """Plan prices to be replaced with additional prices on the subscription.

    (Only available for accounts that have migrated off of legacy subscription
    overrides)
    """

    start_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    trial_duration_days: Optional[int]
    """The duration of the trial period in days.

    If not provided, this defaults to the value specified in the plan. If `0` is
    provided, the trial on the plan will be skipped.
    """

    usage_customer_ids: Optional[SequenceNotStr[str]]
    """
    A list of customer IDs whose usage events will be aggregated and billed under
    this subscription. By default, a subscription only considers usage events
    associated with its attached customer's customer_id. When usage_customer_ids is
    provided, the subscription includes usage events from the specified customers
    only. Provided usage_customer_ids must be either the customer for this
    subscription itself, or any of that customer's children.
    """


AddAdjustmentAdjustment: TypeAlias = Union[
    NewPercentageDiscount, NewUsageDiscount, NewAmountDiscount, NewMinimum, NewMaximum
]


class AddAdjustment(TypedDict, total=False):
    adjustment: Required[AddAdjustmentAdjustment]
    """The definition of a new adjustment to create and add to the subscription."""

    end_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The end date of the adjustment interval.

    This is the date that the adjustment will stop affecting prices on the
    subscription.
    """

    plan_phase_order: Optional[int]
    """The phase to add this adjustment to."""

    start_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The start date of the adjustment interval.

    This is the date that the adjustment will start affecting prices on the
    subscription. If null, the adjustment will start when the phase or subscription
    starts.
    """


class AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier(TypedDict, total=False):
    tier_lower_bound: Required[str]
    """Inclusive tier starting value"""

    unit_amount: Required[str]
    """Amount per unit"""


class AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig(TypedDict, total=False):
    tiers: Required[Iterable[AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier]]
    """
    Tiers for rating based on total usage quantities into the specified tier with
    proration
    """


AddPricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig: TypeAlias = Union[
    UnitConversionRateConfig, TieredConversionRateConfig
]


class AddPricePriceNewSubscriptionTieredWithProrationPrice(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["tiered_with_proration"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    tiered_with_proration_config: Required[
        AddPricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig
    ]
    """Configuration for tiered_with_proration pricing"""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[AddPricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig]
    """The configuration for the rate of the price currency to the invoicing currency."""

    currency: Optional[str]
    """
    An ISO 4217 currency string, or custom pricing unit identifier, in which this
    price is billed.
    """

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    reference_id: Optional[str]
    """
    A transient ID that can be used to reference this price when adding adjustments
    in the same API call.
    """


class AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig(
    TypedDict, total=False
):
    grouping_key: Required[str]
    """The event property used to group before applying thresholds"""

    maximum_charge: Required[str]
    """The maximum amount to charge each group"""

    minimum_charge: Required[str]
    """The minimum amount to charge each group, regardless of usage"""

    per_unit_rate: Required[str]
    """The base price charged per group"""


AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig: TypeAlias = Union[
    UnitConversionRateConfig, TieredConversionRateConfig
]


class AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    grouped_with_min_max_thresholds_config: Required[
        AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig
    ]
    """Configuration for grouped_with_min_max_thresholds pricing"""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["grouped_with_min_max_thresholds"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig]
    """The configuration for the rate of the price currency to the invoicing currency."""

    currency: Optional[str]
    """
    An ISO 4217 currency string, or custom pricing unit identifier, in which this
    price is billed.
    """

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    reference_id: Optional[str]
    """
    A transient ID that can be used to reference this price when adding adjustments
    in the same API call.
    """


AddPricePrice: TypeAlias = Union[
    NewSubscriptionUnitPriceParam,
    NewSubscriptionTieredPriceParam,
    NewSubscriptionBulkPriceParam,
    NewSubscriptionPackagePriceParam,
    NewSubscriptionMatrixPriceParam,
    NewSubscriptionThresholdTotalAmountPriceParam,
    NewSubscriptionTieredPackagePriceParam,
    NewSubscriptionTieredWithMinimumPriceParam,
    NewSubscriptionGroupedTieredPriceParam,
    NewSubscriptionTieredPackageWithMinimumPriceParam,
    NewSubscriptionPackageWithAllocationPriceParam,
    NewSubscriptionUnitWithPercentPriceParam,
    NewSubscriptionMatrixWithAllocationPriceParam,
    AddPricePriceNewSubscriptionTieredWithProrationPrice,
    NewSubscriptionUnitWithProrationPriceParam,
    NewSubscriptionGroupedAllocationPriceParam,
    NewSubscriptionBulkWithProrationPriceParam,
    NewSubscriptionGroupedWithProratedMinimumPriceParam,
    NewSubscriptionGroupedWithMeteredMinimumPriceParam,
    AddPricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice,
    NewSubscriptionMatrixWithDisplayNamePriceParam,
    NewSubscriptionGroupedTieredPackagePriceParam,
    NewSubscriptionMaxGroupTieredPackagePriceParam,
    NewSubscriptionScalableMatrixWithUnitPricingPriceParam,
    NewSubscriptionScalableMatrixWithTieredPricingPriceParam,
    NewSubscriptionCumulativeGroupedBulkPriceParam,
    NewSubscriptionMinimumCompositePriceParam,
]


class AddPrice(TypedDict, total=False):
    allocation_price: Optional[NewAllocationPrice]
    """The definition of a new allocation price to create and add to the subscription."""

    discounts: Optional[Iterable[DiscountOverrideParam]]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's discounts for this price.
    """

    end_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The end date of the price interval.

    This is the date that the price will stop billing on the subscription. If null,
    billing will end when the phase or subscription ends.
    """

    external_price_id: Optional[str]
    """The external price id of the price to add to the subscription."""

    maximum_amount: Optional[str]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's maximum amount for this price.
    """

    minimum_amount: Optional[str]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's minimum amount for this price.
    """

    plan_phase_order: Optional[int]
    """The phase to add this price to."""

    price: Optional[AddPricePrice]
    """New subscription price request body params."""

    price_id: Optional[str]
    """The id of the price to add to the subscription."""

    start_date: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """The start date of the price interval.

    This is the date that the price will start billing on the subscription. If null,
    billing will start when the phase or subscription starts.
    """


class RemoveAdjustment(TypedDict, total=False):
    adjustment_id: Required[str]
    """The id of the adjustment to remove on the subscription."""


class RemovePrice(TypedDict, total=False):
    external_price_id: Optional[str]
    """The external price id of the price to remove on the subscription."""

    price_id: Optional[str]
    """The id of the price to remove on the subscription."""


ReplaceAdjustmentAdjustment: TypeAlias = Union[
    NewPercentageDiscount, NewUsageDiscount, NewAmountDiscount, NewMinimum, NewMaximum
]


class ReplaceAdjustment(TypedDict, total=False):
    adjustment: Required[ReplaceAdjustmentAdjustment]
    """The definition of a new adjustment to create and add to the subscription."""

    replaces_adjustment_id: Required[str]
    """The id of the adjustment on the plan to replace in the subscription."""


class ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier(TypedDict, total=False):
    tier_lower_bound: Required[str]
    """Inclusive tier starting value"""

    unit_amount: Required[str]
    """Amount per unit"""


class ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig(TypedDict, total=False):
    tiers: Required[Iterable[ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfigTier]]
    """
    Tiers for rating based on total usage quantities into the specified tier with
    proration
    """


ReplacePricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig: TypeAlias = Union[
    UnitConversionRateConfig, TieredConversionRateConfig
]


class ReplacePricePriceNewSubscriptionTieredWithProrationPrice(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["tiered_with_proration"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    tiered_with_proration_config: Required[
        ReplacePricePriceNewSubscriptionTieredWithProrationPriceTieredWithProrationConfig
    ]
    """Configuration for tiered_with_proration pricing"""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[ReplacePricePriceNewSubscriptionTieredWithProrationPriceConversionRateConfig]
    """The configuration for the rate of the price currency to the invoicing currency."""

    currency: Optional[str]
    """
    An ISO 4217 currency string, or custom pricing unit identifier, in which this
    price is billed.
    """

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    reference_id: Optional[str]
    """
    A transient ID that can be used to reference this price when adding adjustments
    in the same API call.
    """


class ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig(
    TypedDict, total=False
):
    grouping_key: Required[str]
    """The event property used to group before applying thresholds"""

    maximum_charge: Required[str]
    """The maximum amount to charge each group"""

    minimum_charge: Required[str]
    """The minimum amount to charge each group, regardless of usage"""

    per_unit_rate: Required[str]
    """The base price charged per group"""


ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig: TypeAlias = Union[
    UnitConversionRateConfig, TieredConversionRateConfig
]


class ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    grouped_with_min_max_thresholds_config: Required[
        ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig
    ]
    """Configuration for grouped_with_min_max_thresholds pricing"""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["grouped_with_min_max_thresholds"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[
        ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPriceConversionRateConfig
    ]
    """The configuration for the rate of the price currency to the invoicing currency."""

    currency: Optional[str]
    """
    An ISO 4217 currency string, or custom pricing unit identifier, in which this
    price is billed.
    """

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    reference_id: Optional[str]
    """
    A transient ID that can be used to reference this price when adding adjustments
    in the same API call.
    """


ReplacePricePrice: TypeAlias = Union[
    NewSubscriptionUnitPriceParam,
    NewSubscriptionTieredPriceParam,
    NewSubscriptionBulkPriceParam,
    NewSubscriptionPackagePriceParam,
    NewSubscriptionMatrixPriceParam,
    NewSubscriptionThresholdTotalAmountPriceParam,
    NewSubscriptionTieredPackagePriceParam,
    NewSubscriptionTieredWithMinimumPriceParam,
    NewSubscriptionGroupedTieredPriceParam,
    NewSubscriptionTieredPackageWithMinimumPriceParam,
    NewSubscriptionPackageWithAllocationPriceParam,
    NewSubscriptionUnitWithPercentPriceParam,
    NewSubscriptionMatrixWithAllocationPriceParam,
    ReplacePricePriceNewSubscriptionTieredWithProrationPrice,
    NewSubscriptionUnitWithProrationPriceParam,
    NewSubscriptionGroupedAllocationPriceParam,
    NewSubscriptionBulkWithProrationPriceParam,
    NewSubscriptionGroupedWithProratedMinimumPriceParam,
    NewSubscriptionGroupedWithMeteredMinimumPriceParam,
    ReplacePricePriceNewSubscriptionGroupedWithMinMaxThresholdsPrice,
    NewSubscriptionMatrixWithDisplayNamePriceParam,
    NewSubscriptionGroupedTieredPackagePriceParam,
    NewSubscriptionMaxGroupTieredPackagePriceParam,
    NewSubscriptionScalableMatrixWithUnitPricingPriceParam,
    NewSubscriptionScalableMatrixWithTieredPricingPriceParam,
    NewSubscriptionCumulativeGroupedBulkPriceParam,
    NewSubscriptionMinimumCompositePriceParam,
]


class ReplacePrice(TypedDict, total=False):
    replaces_price_id: Required[str]
    """The id of the price on the plan to replace in the subscription."""

    allocation_price: Optional[NewAllocationPrice]
    """The definition of a new allocation price to create and add to the subscription."""

    discounts: Optional[Iterable[DiscountOverrideParam]]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's discounts for the replacement price.
    """

    external_price_id: Optional[str]
    """The external price id of the price to add to the subscription."""

    fixed_price_quantity: Optional[float]
    """The new quantity of the price, if the price is a fixed price."""

    maximum_amount: Optional[str]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's maximum amount for the replacement price.
    """

    minimum_amount: Optional[str]
    """[DEPRECATED] Use add_adjustments instead.

    The subscription's minimum amount for the replacement price.
    """

    price: Optional[ReplacePricePrice]
    """New subscription price request body params."""

    price_id: Optional[str]
    """The id of the price to add to the subscription."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from typing_extensions import Literal, Required, TypeAlias, TypedDict

from .shared_params.unit_conversion_rate_config import UnitConversionRateConfig
from .shared_params.tiered_conversion_rate_config import TieredConversionRateConfig
from .shared_params.new_billing_cycle_configuration import NewBillingCycleConfiguration
from .shared_params.new_dimensional_price_configuration import NewDimensionalPriceConfiguration

__all__ = [
    "NewSubscriptionTieredPackagePriceParam",
    "TieredPackageConfig",
    "TieredPackageConfigTier",
    "ConversionRateConfig",
]


class TieredPackageConfigTier(TypedDict, total=False):
    per_unit: Required[str]
    """Price per package"""

    tier_lower_bound: Required[str]
    """Tier lower bound"""


class TieredPackageConfig(TypedDict, total=False):
    package_size: Required[str]
    """Package size"""

    tiers: Required[Iterable[TieredPackageConfigTier]]
    """Apply tiered pricing after rounding up the quantity to the package size.

    Tiers are defined using exclusive lower bounds. The tier bounds are defined
    based on the total quantity rather than the number of packages, so they must be
    multiples of the package size.
    """


ConversionRateConfig: TypeAlias = Union[UnitConversionRateConfig, TieredConversionRateConfig]


class NewSubscriptionTieredPackagePriceParam(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["tiered_package"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    tiered_package_config: Required[TieredPackageConfig]
    """Configuration for tiered_package pricing"""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[ConversionRateConfig]
    """The configuration for the rate of the price currency to the invoicing currency."""

    currency: Optional[str]
    """
    An ISO 4217 currency string, or custom pricing unit identifier, in which this
    price is billed.
    """

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """

    reference_id: Optional[str]
    """
    A transient ID that can be used to reference this price when adding adjustments
    in the same API call.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

__all__ = ["SubscriptionChangeApplyParams"]


class SubscriptionChangeApplyParams(TypedDict, total=False):
    description: Optional[str]
    """Description to apply to the balance transaction representing this credit."""

    previously_collected_amount: Optional[str]
    """Amount already collected to apply to the customer's balance."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from datetime import datetime

from ..._models import BaseModel
from .per_price_cost import PerPriceCost

__all__ = ["AggregatedCost"]


class AggregatedCost(BaseModel):
    per_price_costs: List[PerPriceCost]

    subtotal: str
    """Total costs for the timeframe, excluding any minimums and discounts."""

    timeframe_end: datetime

    timeframe_start: datetime

    total: str
    """Total costs for the timeframe, including any minimums and discounts."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from .._models import BaseModel
from .evaluate_price_group import EvaluatePriceGroup

__all__ = ["PriceEvaluatePreviewEventsResponse", "Data"]


class Data(BaseModel):
    currency: str
    """The currency of the price"""

    price_groups: List[EvaluatePriceGroup]
    """The computed price groups associated with input price."""

    external_price_id: Optional[str] = None
    """The external ID of the price"""

    inline_price_index: Optional[int] = None
    """The index of the inline price"""

    price_id: Optional[str] = None
    """The ID of the price"""


class PriceEvaluatePreviewEventsResponse(BaseModel):
    data: List[Data]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..._models import BaseModel

__all__ = ["ConversionRateUnitConfig"]


class ConversionRateUnitConfig(BaseModel):
    unit_amount: str
    """Amount per unit of overage"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from orb import Orb, AsyncOrb
from orb.types import (
    EventIngestResponse,
    EventSearchResponse,
    EventUpdateResponse,
    EventDeprecateResponse,
)
from orb._utils import parse_datetime
from tests.utils import assert_matches_type

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestEvents:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_update(self, client: Orb) -> None:
        event = client.events.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        )
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: Orb) -> None:
        event = client.events.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
            customer_id="customer_id",
            external_customer_id="external_customer_id",
        )
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: Orb) -> None:
        response = client.events.with_raw_response.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: Orb) -> None:
        with client.events.with_streaming_response.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = response.parse()
            assert_matches_type(EventUpdateResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `event_id` but received ''"):
            client.events.with_raw_response.update(
                event_id="",
                event_name="event_name",
                properties={"foo": "bar"},
                timestamp=parse_datetime("2020-12-09T16:09:53Z"),
            )

    @parametrize
    def test_method_deprecate(self, client: Orb) -> None:
        event = client.events.deprecate(
            "event_id",
        )
        assert_matches_type(EventDeprecateResponse, event, path=["response"])

    @parametrize
    def test_raw_response_deprecate(self, client: Orb) -> None:
        response = client.events.with_raw_response.deprecate(
            "event_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventDeprecateResponse, event, path=["response"])

    @parametrize
    def test_streaming_response_deprecate(self, client: Orb) -> None:
        with client.events.with_streaming_response.deprecate(
            "event_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = response.parse()
            assert_matches_type(EventDeprecateResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_deprecate(self, client: Orb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `event_id` but received ''"):
            client.events.with_raw_response.deprecate(
                "",
            )

    @parametrize
    def test_method_ingest(self, client: Orb) -> None:
        event = client.events.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        )
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    def test_method_ingest_with_all_params(self, client: Orb) -> None:
        event = client.events.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                    "customer_id": "customer_id",
                    "external_customer_id": "external_customer_id",
                }
            ],
            backfill_id="backfill_id",
            debug=True,
        )
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    def test_raw_response_ingest(self, client: Orb) -> None:
        response = client.events.with_raw_response.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    def test_streaming_response_ingest(self, client: Orb) -> None:
        with client.events.with_streaming_response.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = response.parse()
            assert_matches_type(EventIngestResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_search(self, client: Orb) -> None:
        event = client.events.search(
            event_ids=["string"],
        )
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    def test_method_search_with_all_params(self, client: Orb) -> None:
        event = client.events.search(
            event_ids=["string"],
            timeframe_end=parse_datetime("2019-12-27T18:11:19.117Z"),
            timeframe_start=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    def test_raw_response_search(self, client: Orb) -> None:
        response = client.events.with_raw_response.search(
            event_ids=["string"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    def test_streaming_response_search(self, client: Orb) -> None:
        with client.events.with_streaming_response.search(
            event_ids=["string"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = response.parse()
            assert_matches_type(EventSearchResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncEvents:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_update(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        )
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
            customer_id="customer_id",
            external_customer_id="external_customer_id",
        )
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncOrb) -> None:
        response = await async_client.events.with_raw_response.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventUpdateResponse, event, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncOrb) -> None:
        async with async_client.events.with_streaming_response.update(
            event_id="event_id",
            event_name="event_name",
            properties={"foo": "bar"},
            timestamp=parse_datetime("2020-12-09T16:09:53Z"),
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = await response.parse()
            assert_matches_type(EventUpdateResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `event_id` but received ''"):
            await async_client.events.with_raw_response.update(
                event_id="",
                event_name="event_name",
                properties={"foo": "bar"},
                timestamp=parse_datetime("2020-12-09T16:09:53Z"),
            )

    @parametrize
    async def test_method_deprecate(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.deprecate(
            "event_id",
        )
        assert_matches_type(EventDeprecateResponse, event, path=["response"])

    @parametrize
    async def test_raw_response_deprecate(self, async_client: AsyncOrb) -> None:
        response = await async_client.events.with_raw_response.deprecate(
            "event_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventDeprecateResponse, event, path=["response"])

    @parametrize
    async def test_streaming_response_deprecate(self, async_client: AsyncOrb) -> None:
        async with async_client.events.with_streaming_response.deprecate(
            "event_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = await response.parse()
            assert_matches_type(EventDeprecateResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_deprecate(self, async_client: AsyncOrb) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `event_id` but received ''"):
            await async_client.events.with_raw_response.deprecate(
                "",
            )

    @parametrize
    async def test_method_ingest(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        )
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    async def test_method_ingest_with_all_params(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                    "customer_id": "customer_id",
                    "external_customer_id": "external_customer_id",
                }
            ],
            backfill_id="backfill_id",
            debug=True,
        )
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    async def test_raw_response_ingest(self, async_client: AsyncOrb) -> None:
        response = await async_client.events.with_raw_response.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventIngestResponse, event, path=["response"])

    @parametrize
    async def test_streaming_response_ingest(self, async_client: AsyncOrb) -> None:
        async with async_client.events.with_streaming_response.ingest(
            events=[
                {
                    "event_name": "event_name",
                    "idempotency_key": "idempotency_key",
                    "properties": {"foo": "bar"},
                    "timestamp": parse_datetime("2020-12-09T16:09:53Z"),
                }
            ],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = await response.parse()
            assert_matches_type(EventIngestResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_search(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.search(
            event_ids=["string"],
        )
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    async def test_method_search_with_all_params(self, async_client: AsyncOrb) -> None:
        event = await async_client.events.search(
            event_ids=["string"],
            timeframe_end=parse_datetime("2019-12-27T18:11:19.117Z"),
            timeframe_start=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    async def test_raw_response_search(self, async_client: AsyncOrb) -> None:
        response = await async_client.events.with_raw_response.search(
            event_ids=["string"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        event = response.parse()
        assert_matches_type(EventSearchResponse, event, path=["response"])

    @parametrize
    async def test_streaming_response_search(self, async_client: AsyncOrb) -> None:
        async with async_client.events.with_streaming_response.search(
            event_ids=["string"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            event = await response.parse()
            assert_matches_type(EventSearchResponse, event, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .external_plan_id_create_plan_version_params import (
    ExternalPlanIDCreatePlanVersionParams as ExternalPlanIDCreatePlanVersionParams,
)
from .external_plan_id_set_default_plan_version_params import (
    ExternalPlanIDSetDefaultPlanVersionParams as ExternalPlanIDSetDefaultPlanVersionParams,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Annotated, TypeAlias

from ..._utils import PropertyInfo
from .trial_discount import TrialDiscount
from .usage_discount import UsageDiscount
from .amount_discount import AmountDiscount
from .percentage_discount import PercentageDiscount

__all__ = ["Discount"]

Discount: TypeAlias = Annotated[
    Union[PercentageDiscount, TrialDiscount, UsageDiscount, AmountDiscount], PropertyInfo(discriminator="discount_type")
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from ..._models import BaseModel

__all__ = ["MatrixWithAllocationConfig", "MatrixValue"]


class MatrixValue(BaseModel):
    dimension_values: List[Optional[str]]
    """One or two matrix keys to filter usage to this Matrix value by.

    For example, ["region", "tier"] could be used to filter cloud usage by a cloud
    region and an instance tier.
    """

    unit_amount: str
    """Unit price for the specified dimension_values"""


class MatrixWithAllocationConfig(BaseModel):
    allocation: str
    """Usage allocation"""

    default_unit_amount: str
    """Default per unit rate for any usage not bucketed into a specified matrix_value"""

    dimensions: List[Optional[str]]
    """One or two event property values to evaluate matrix groups by"""

    matrix_values: List[MatrixValue]
    """Matrix values configuration"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Mapping
from typing_extensions import Self, override

import httpx

from . import _exceptions
from ._qs import Querystring
from ._types import (
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
    not_given,
)
from ._utils import (
    is_given,
    is_mapping,
    get_async_library,
)
from ._compat import cached_property
from ._version import __version__
from ._streaming import Stream as Stream, AsyncStream as AsyncStream
from ._exceptions import OrbError, APIStatusError
from ._base_client import (
    DEFAULT_MAX_RETRIES,
    SyncAPIClient,
    AsyncAPIClient,
)

if TYPE_CHECKING:
    from .resources import (
        beta,
        items,
        plans,
        alerts,
        events,
        prices,
        coupons,
        metrics,
        invoices,
        webhooks,
        customers,
        top_level,
        credit_notes,
        subscriptions,
        invoice_line_items,
        subscription_changes,
        dimensional_price_groups,
    )
    from .resources.items import Items, AsyncItems
    from .resources.alerts import Alerts, AsyncAlerts
    from .resources.metrics import Metrics, AsyncMetrics
    from .resources.invoices import Invoices, AsyncInvoices
    from .resources.beta.beta import Beta, AsyncBeta
    from .resources.top_level import TopLevel, AsyncTopLevel
    from .resources.plans.plans import Plans, AsyncPlans
    from .resources.credit_notes import CreditNotes, AsyncCreditNotes
    from .resources.events.events import Events, AsyncEvents
    from .resources.prices.prices import Prices, AsyncPrices
    from .resources.subscriptions import Subscriptions, AsyncSubscriptions
    from .resources.coupons.coupons import Coupons, AsyncCoupons
    from .resources.invoice_line_items import InvoiceLineItems, AsyncInvoiceLineItems
    from .resources.customers.customers import Customers, AsyncCustomers
    from .resources.subscription_changes import SubscriptionChanges, AsyncSubscriptionChanges
    from .resources.dimensional_price_groups.dimensional_price_groups import (
        DimensionalPriceGroups,
        AsyncDimensionalPriceGroups,
    )

__all__ = ["Timeout", "Transport", "ProxiesTypes", "RequestOptions", "Orb", "AsyncOrb", "Client", "AsyncClient"]


class Orb(SyncAPIClient):
    # client options
    api_key: str
    webhook_secret: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous Orb client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `ORB_API_KEY`
        - `webhook_secret` from `ORB_WEBHOOK_SECRET`
        """
        if api_key is None:
            api_key = os.environ.get("ORB_API_KEY")
        if api_key is None:
            raise OrbError(
                "The api_key client option must be set either by passing api_key to the client or by setting the ORB_API_KEY environment variable"
            )
        self.api_key = api_key

        if webhook_secret is None:
            webhook_secret = os.environ.get("ORB_WEBHOOK_SECRET")
        self.webhook_secret = webhook_secret

        if base_url is None:
            base_url = os.environ.get("ORB_BASE_URL")
        if base_url is None:
            base_url = f"https://api.withorb.com/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._idempotency_header = "Idempotency-Key"

    @cached_property
    def top_level(self) -> TopLevel:
        from .resources.top_level import TopLevel

        return TopLevel(self)

    @cached_property
    def beta(self) -> Beta:
        from .resources.beta import Beta

        return Beta(self)

    @cached_property
    def coupons(self) -> Coupons:
        from .resources.coupons import Coupons

        return Coupons(self)

    @cached_property
    def credit_notes(self) -> CreditNotes:
        from .resources.credit_notes import CreditNotes

        return CreditNotes(self)

    @cached_property
    def customers(self) -> Customers:
        from .resources.customers import Customers

        return Customers(self)

    @cached_property
    def events(self) -> Events:
        from .resources.events import Events

        return Events(self)

    @cached_property
    def invoice_line_items(self) -> InvoiceLineItems:
        from .resources.invoice_line_items import InvoiceLineItems

        return InvoiceLineItems(self)

    @cached_property
    def invoices(self) -> Invoices:
        from .resources.invoices import Invoices

        return Invoices(self)

    @cached_property
    def items(self) -> Items:
        from .resources.items import Items

        return Items(self)

    @cached_property
    def metrics(self) -> Metrics:
        from .resources.metrics import Metrics

        return Metrics(self)

    @cached_property
    def plans(self) -> Plans:
        from .resources.plans import Plans

        return Plans(self)

    @cached_property
    def prices(self) -> Prices:
        from .resources.prices import Prices

        return Prices(self)

    @cached_property
    def subscriptions(self) -> Subscriptions:
        from .resources.subscriptions import Subscriptions

        return Subscriptions(self)

    @cached_property
    def alerts(self) -> Alerts:
        from .resources.alerts import Alerts

        return Alerts(self)

    @cached_property
    def dimensional_price_groups(self) -> DimensionalPriceGroups:
        from .resources.dimensional_price_groups import DimensionalPriceGroups

        return DimensionalPriceGroups(self)

    @cached_property
    def subscription_changes(self) -> SubscriptionChanges:
        from .resources.subscription_changes import SubscriptionChanges

        return SubscriptionChanges(self)

    @cached_property
    def webhooks(self) -> webhooks.Webhooks:
        from .resources.webhooks import Webhooks

        return Webhooks(self)

    @cached_property
    def with_raw_response(self) -> OrbWithRawResponse:
        return OrbWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> OrbWithStreamedResponse:
        return OrbWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="brackets")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": "false",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            webhook_secret=webhook_secret or self.webhook_secret,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        type_ = body.get("type") if is_mapping(body) else None
        if type_ == "https://docs.withorb.com/reference/error-responses#400-constraint-violation":
            return _exceptions.ConstraintViolation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-duplicate-resource-creation":
            return _exceptions.DuplicateResourceCreation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-feature-not-available":
            return _exceptions.FeatureNotAvailable(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-request-validation-errors":
            return _exceptions.RequestValidationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#401-authentication-error":
            return _exceptions.OrbAuthenticationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-resource-not-found":
            return _exceptions.ResourceNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-url-not-found":
            return _exceptions.URLNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#409-resource-conflict":
            return _exceptions.ResourceConflict(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-request-too-large":
            return _exceptions.RequestTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-resource-too-large":
            return _exceptions.ResourceTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#429-too-many-requests":
            return _exceptions.TooManyRequests(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#500-internal-server-error":
            return _exceptions.OrbInternalServerError(err_msg, response=response, body=body)
        if response.status_code >= 500:
            return _exceptions.OrbInternalServerError(
                err_msg,
                response=response,
                body={
                    "status": 500,
                    "type": "https://docs.withorb.com/reference/error-responses#500-internal-server-error",
                    "detail": None,
                    "title": None,
                },
            )

        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class AsyncOrb(AsyncAPIClient):
    # client options
    api_key: str
    webhook_secret: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client.
        # We provide a `DefaultAsyncHttpxClient` class that you can pass to retain the default values we use for `limits`, `timeout` & `follow_redirects`.
        # See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async AsyncOrb client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `ORB_API_KEY`
        - `webhook_secret` from `ORB_WEBHOOK_SECRET`
        """
        if api_key is None:
            api_key = os.environ.get("ORB_API_KEY")
        if api_key is None:
            raise OrbError(
                "The api_key client option must be set either by passing api_key to the client or by setting the ORB_API_KEY environment variable"
            )
        self.api_key = api_key

        if webhook_secret is None:
            webhook_secret = os.environ.get("ORB_WEBHOOK_SECRET")
        self.webhook_secret = webhook_secret

        if base_url is None:
            base_url = os.environ.get("ORB_BASE_URL")
        if base_url is None:
            base_url = f"https://api.withorb.com/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._idempotency_header = "Idempotency-Key"

    @cached_property
    def top_level(self) -> AsyncTopLevel:
        from .resources.top_level import AsyncTopLevel

        return AsyncTopLevel(self)

    @cached_property
    def beta(self) -> AsyncBeta:
        from .resources.beta import AsyncBeta

        return AsyncBeta(self)

    @cached_property
    def coupons(self) -> AsyncCoupons:
        from .resources.coupons import AsyncCoupons

        return AsyncCoupons(self)

    @cached_property
    def credit_notes(self) -> AsyncCreditNotes:
        from .resources.credit_notes import AsyncCreditNotes

        return AsyncCreditNotes(self)

    @cached_property
    def customers(self) -> AsyncCustomers:
        from .resources.customers import AsyncCustomers

        return AsyncCustomers(self)

    @cached_property
    def events(self) -> AsyncEvents:
        from .resources.events import AsyncEvents

        return AsyncEvents(self)

    @cached_property
    def invoice_line_items(self) -> AsyncInvoiceLineItems:
        from .resources.invoice_line_items import AsyncInvoiceLineItems

        return AsyncInvoiceLineItems(self)

    @cached_property
    def invoices(self) -> AsyncInvoices:
        from .resources.invoices import AsyncInvoices

        return AsyncInvoices(self)

    @cached_property
    def items(self) -> AsyncItems:
        from .resources.items import AsyncItems

        return AsyncItems(self)

    @cached_property
    def metrics(self) -> AsyncMetrics:
        from .resources.metrics import AsyncMetrics

        return AsyncMetrics(self)

    @cached_property
    def plans(self) -> AsyncPlans:
        from .resources.plans import AsyncPlans

        return AsyncPlans(self)

    @cached_property
    def prices(self) -> AsyncPrices:
        from .resources.prices import AsyncPrices

        return AsyncPrices(self)

    @cached_property
    def subscriptions(self) -> AsyncSubscriptions:
        from .resources.subscriptions import AsyncSubscriptions

        return AsyncSubscriptions(self)

    @cached_property
    def alerts(self) -> AsyncAlerts:
        from .resources.alerts import AsyncAlerts

        return AsyncAlerts(self)

    @cached_property
    def dimensional_price_groups(self) -> AsyncDimensionalPriceGroups:
        from .resources.dimensional_price_groups import AsyncDimensionalPriceGroups

        return AsyncDimensionalPriceGroups(self)

    @cached_property
    def subscription_changes(self) -> AsyncSubscriptionChanges:
        from .resources.subscription_changes import AsyncSubscriptionChanges

        return AsyncSubscriptionChanges(self)

    @cached_property
    def webhooks(self) -> webhooks.AsyncWebhooks:
        from .resources.webhooks import AsyncWebhooks

        return AsyncWebhooks(self)

    @cached_property
    def with_raw_response(self) -> AsyncOrbWithRawResponse:
        return AsyncOrbWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncOrbWithStreamedResponse:
        return AsyncOrbWithStreamedResponse(self)

    @property
    @override
    def qs(self) -> Querystring:
        return Querystring(array_format="brackets")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:
        return {
            **super().default_headers,
            "X-Stainless-Async": f"async:{get_async_library()}",
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        webhook_secret: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = not_given,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = not_given,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
        _extra_kwargs: Mapping[str, Any] = {},
    ) -> Self:
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError("The `default_headers` and `set_default_headers` arguments are mutually exclusive")

        if default_query is not None and set_default_query is not None:
            raise ValueError("The `default_query` and `set_default_query` arguments are mutually exclusive")

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            webhook_secret=webhook_secret or self.webhook_secret,
            base_url=base_url or self.base_url,
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
            **_extra_kwargs,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:
        type_ = body.get("type") if is_mapping(body) else None
        if type_ == "https://docs.withorb.com/reference/error-responses#400-constraint-violation":
            return _exceptions.ConstraintViolation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-duplicate-resource-creation":
            return _exceptions.DuplicateResourceCreation(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-feature-not-available":
            return _exceptions.FeatureNotAvailable(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#400-request-validation-errors":
            return _exceptions.RequestValidationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#401-authentication-error":
            return _exceptions.OrbAuthenticationError(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-resource-not-found":
            return _exceptions.ResourceNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#404-url-not-found":
            return _exceptions.URLNotFound(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#409-resource-conflict":
            return _exceptions.ResourceConflict(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-request-too-large":
            return _exceptions.RequestTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#413-resource-too-large":
            return _exceptions.ResourceTooLarge(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#429-too-many-requests":
            return _exceptions.TooManyRequests(err_msg, response=response, body=body)

        if type_ == "https://docs.withorb.com/reference/error-responses#500-internal-server-error":
            return _exceptions.OrbInternalServerError(err_msg, response=response, body=body)
        if response.status_code >= 500:
            return _exceptions.OrbInternalServerError(
                err_msg,
                response=response,
                body={
                    "status": 500,
                    "type": "https://docs.withorb.com/reference/error-responses#500-internal-server-error",
                    "detail": None,
                    "title": None,
                },
            )

        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(err_msg, response=response, body=body)

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(err_msg, response=response, body=body)

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)


class OrbWithRawResponse:
    _client: Orb

    def __init__(self, client: Orb) -> None:
        self._client = client

    @cached_property
    def top_level(self) -> top_level.TopLevelWithRawResponse:
        from .resources.top_level import TopLevelWithRawResponse

        return TopLevelWithRawResponse(self._client.top_level)

    @cached_property
    def beta(self) -> beta.BetaWithRawResponse:
        from .resources.beta import BetaWithRawResponse

        return BetaWithRawResponse(self._client.beta)

    @cached_property
    def coupons(self) -> coupons.CouponsWithRawResponse:
        from .resources.coupons import CouponsWithRawResponse

        return CouponsWithRawResponse(self._client.coupons)

    @cached_property
    def credit_notes(self) -> credit_notes.CreditNotesWithRawResponse:
        from .resources.credit_notes import CreditNotesWithRawResponse

        return CreditNotesWithRawResponse(self._client.credit_notes)

    @cached_property
    def customers(self) -> customers.CustomersWithRawResponse:
        from .resources.customers import CustomersWithRawResponse

        return CustomersWithRawResponse(self._client.customers)

    @cached_property
    def events(self) -> events.EventsWithRawResponse:
        from .resources.events import EventsWithRawResponse

        return EventsWithRawResponse(self._client.events)

    @cached_property
    def invoice_line_items(self) -> invoice_line_items.InvoiceLineItemsWithRawResponse:
        from .resources.invoice_line_items import InvoiceLineItemsWithRawResponse

        return InvoiceLineItemsWithRawResponse(self._client.invoice_line_items)

    @cached_property
    def invoices(self) -> invoices.InvoicesWithRawResponse:
        from .resources.invoices import InvoicesWithRawResponse

        return InvoicesWithRawResponse(self._client.invoices)

    @cached_property
    def items(self) -> items.ItemsWithRawResponse:
        from .resources.items import ItemsWithRawResponse

        return ItemsWithRawResponse(self._client.items)

    @cached_property
    def metrics(self) -> metrics.MetricsWithRawResponse:
        from .resources.metrics import MetricsWithRawResponse

        return MetricsWithRawResponse(self._client.metrics)

    @cached_property
    def plans(self) -> plans.PlansWithRawResponse:
        from .resources.plans import PlansWithRawResponse

        return PlansWithRawResponse(self._client.plans)

    @cached_property
    def prices(self) -> prices.PricesWithRawResponse:
        from .resources.prices import PricesWithRawResponse

        return PricesWithRawResponse(self._client.prices)

    @cached_property
    def subscriptions(self) -> subscriptions.SubscriptionsWithRawResponse:
        from .resources.subscriptions import SubscriptionsWithRawResponse

        return SubscriptionsWithRawResponse(self._client.subscriptions)

    @cached_property
    def alerts(self) -> alerts.AlertsWithRawResponse:
        from .resources.alerts import AlertsWithRawResponse

        return AlertsWithRawResponse(self._client.alerts)

    @cached_property
    def dimensional_price_groups(self) -> dimensional_price_groups.DimensionalPriceGroupsWithRawResponse:
        from .resources.dimensional_price_groups import DimensionalPriceGroupsWithRawResponse

        return DimensionalPriceGroupsWithRawResponse(self._client.dimensional_price_groups)

    @cached_property
    def subscription_changes(self) -> subscription_changes.SubscriptionChangesWithRawResponse:
        from .resources.subscription_changes import SubscriptionChangesWithRawResponse

        return SubscriptionChangesWithRawResponse(self._client.subscription_changes)


class AsyncOrbWithRawResponse:
    _client: AsyncOrb

    def __init__(self, client: AsyncOrb) -> None:
        self._client = client

    @cached_property
    def top_level(self) -> top_level.AsyncTopLevelWithRawResponse:
        from .resources.top_level import AsyncTopLevelWithRawResponse

        return AsyncTopLevelWithRawResponse(self._client.top_level)

    @cached_property
    def beta(self) -> beta.AsyncBetaWithRawResponse:
        from .resources.beta import AsyncBetaWithRawResponse

        return AsyncBetaWithRawResponse(self._client.beta)

    @cached_property
    def coupons(self) -> coupons.AsyncCouponsWithRawResponse:
        from .resources.coupons import AsyncCouponsWithRawResponse

        return AsyncCouponsWithRawResponse(self._client.coupons)

    @cached_property
    def credit_notes(self) -> credit_notes.AsyncCreditNotesWithRawResponse:
        from .resources.credit_notes import AsyncCreditNotesWithRawResponse

        return AsyncCreditNotesWithRawResponse(self._client.credit_notes)

    @cached_property
    def customers(self) -> customers.AsyncCustomersWithRawResponse:
        from .resources.customers import AsyncCustomersWithRawResponse

        return AsyncCustomersWithRawResponse(self._client.customers)

    @cached_property
    def events(self) -> events.AsyncEventsWithRawResponse:
        from .resources.events import AsyncEventsWithRawResponse

        return AsyncEventsWithRawResponse(self._client.events)

    @cached_property
    def invoice_line_items(self) -> invoice_line_items.AsyncInvoiceLineItemsWithRawResponse:
        from .resources.invoice_line_items import AsyncInvoiceLineItemsWithRawResponse

        return AsyncInvoiceLineItemsWithRawResponse(self._client.invoice_line_items)

    @cached_property
    def invoices(self) -> invoices.AsyncInvoicesWithRawResponse:
        from .resources.invoices import AsyncInvoicesWithRawResponse

        return AsyncInvoicesWithRawResponse(self._client.invoices)

    @cached_property
    def items(self) -> items.AsyncItemsWithRawResponse:
        from .resources.items import AsyncItemsWithRawResponse

        return AsyncItemsWithRawResponse(self._client.items)

    @cached_property
    def metrics(self) -> metrics.AsyncMetricsWithRawResponse:
        from .resources.metrics import AsyncMetricsWithRawResponse

        return AsyncMetricsWithRawResponse(self._client.metrics)

    @cached_property
    def plans(self) -> plans.AsyncPlansWithRawResponse:
        from .resources.plans import AsyncPlansWithRawResponse

        return AsyncPlansWithRawResponse(self._client.plans)

    @cached_property
    def prices(self) -> prices.AsyncPricesWithRawResponse:
        from .resources.prices import AsyncPricesWithRawResponse

        return AsyncPricesWithRawResponse(self._client.prices)

    @cached_property
    def subscriptions(self) -> subscriptions.AsyncSubscriptionsWithRawResponse:
        from .resources.subscriptions import AsyncSubscriptionsWithRawResponse

        return AsyncSubscriptionsWithRawResponse(self._client.subscriptions)

    @cached_property
    def alerts(self) -> alerts.AsyncAlertsWithRawResponse:
        from .resources.alerts import AsyncAlertsWithRawResponse

        return AsyncAlertsWithRawResponse(self._client.alerts)

    @cached_property
    def dimensional_price_groups(self) -> dimensional_price_groups.AsyncDimensionalPriceGroupsWithRawResponse:
        from .resources.dimensional_price_groups import AsyncDimensionalPriceGroupsWithRawResponse

        return AsyncDimensionalPriceGroupsWithRawResponse(self._client.dimensional_price_groups)

    @cached_property
    def subscription_changes(self) -> subscription_changes.AsyncSubscriptionChangesWithRawResponse:
        from .resources.subscription_changes import AsyncSubscriptionChangesWithRawResponse

        return AsyncSubscriptionChangesWithRawResponse(self._client.subscription_changes)


class OrbWithStreamedResponse:
    _client: Orb

    def __init__(self, client: Orb) -> None:
        self._client = client

    @cached_property
    def top_level(self) -> top_level.TopLevelWithStreamingResponse:
        from .resources.top_level import TopLevelWithStreamingResponse

        return TopLevelWithStreamingResponse(self._client.top_level)

    @cached_property
    def beta(self) -> beta.BetaWithStreamingResponse:
        from .resources.beta import BetaWithStreamingResponse

        return BetaWithStreamingResponse(self._client.beta)

    @cached_property
    def coupons(self) -> coupons.CouponsWithStreamingResponse:
        from .resources.coupons import CouponsWithStreamingResponse

        return CouponsWithStreamingResponse(self._client.coupons)

    @cached_property
    def credit_notes(self) -> credit_notes.CreditNotesWithStreamingResponse:
        from .resources.credit_notes import CreditNotesWithStreamingResponse

        return CreditNotesWithStreamingResponse(self._client.credit_notes)

    @cached_property
    def customers(self) -> customers.CustomersWithStreamingResponse:
        from .resources.customers import CustomersWithStreamingResponse

        return CustomersWithStreamingResponse(self._client.customers)

    @cached_property
    def events(self) -> events.EventsWithStreamingResponse:
        from .resources.events import EventsWithStreamingResponse

        return EventsWithStreamingResponse(self._client.events)

    @cached_property
    def invoice_line_items(self) -> invoice_line_items.InvoiceLineItemsWithStreamingResponse:
        from .resources.invoice_line_items import InvoiceLineItemsWithStreamingResponse

        return InvoiceLineItemsWithStreamingResponse(self._client.invoice_line_items)

    @cached_property
    def invoices(self) -> invoices.InvoicesWithStreamingResponse:
        from .resources.invoices import InvoicesWithStreamingResponse

        return InvoicesWithStreamingResponse(self._client.invoices)

    @cached_property
    def items(self) -> items.ItemsWithStreamingResponse:
        from .resources.items import ItemsWithStreamingResponse

        return ItemsWithStreamingResponse(self._client.items)

    @cached_property
    def metrics(self) -> metrics.MetricsWithStreamingResponse:
        from .resources.metrics import MetricsWithStreamingResponse

        return MetricsWithStreamingResponse(self._client.metrics)

    @cached_property
    def plans(self) -> plans.PlansWithStreamingResponse:
        from .resources.plans import PlansWithStreamingResponse

        return PlansWithStreamingResponse(self._client.plans)

    @cached_property
    def prices(self) -> prices.PricesWithStreamingResponse:
        from .resources.prices import PricesWithStreamingResponse

        return PricesWithStreamingResponse(self._client.prices)

    @cached_property
    def subscriptions(self) -> subscriptions.SubscriptionsWithStreamingResponse:
        from .resources.subscriptions import SubscriptionsWithStreamingResponse

        return SubscriptionsWithStreamingResponse(self._client.subscriptions)

    @cached_property
    def alerts(self) -> alerts.AlertsWithStreamingResponse:
        from .resources.alerts import AlertsWithStreamingResponse

        return AlertsWithStreamingResponse(self._client.alerts)

    @cached_property
    def dimensional_price_groups(self) -> dimensional_price_groups.DimensionalPriceGroupsWithStreamingResponse:
        from .resources.dimensional_price_groups import DimensionalPriceGroupsWithStreamingResponse

        return DimensionalPriceGroupsWithStreamingResponse(self._client.dimensional_price_groups)

    @cached_property
    def subscription_changes(self) -> subscription_changes.SubscriptionChangesWithStreamingResponse:
        from .resources.subscription_changes import SubscriptionChangesWithStreamingResponse

        return SubscriptionChangesWithStreamingResponse(self._client.subscription_changes)


class AsyncOrbWithStreamedResponse:
    _client: AsyncOrb

    def __init__(self, client: AsyncOrb) -> None:
        self._client = client

    @cached_property
    def top_level(self) -> top_level.AsyncTopLevelWithStreamingResponse:
        from .resources.top_level import AsyncTopLevelWithStreamingResponse

        return AsyncTopLevelWithStreamingResponse(self._client.top_level)

    @cached_property
    def beta(self) -> beta.AsyncBetaWithStreamingResponse:
        from .resources.beta import AsyncBetaWithStreamingResponse

        return AsyncBetaWithStreamingResponse(self._client.beta)

    @cached_property
    def coupons(self) -> coupons.AsyncCouponsWithStreamingResponse:
        from .resources.coupons import AsyncCouponsWithStreamingResponse

        return AsyncCouponsWithStreamingResponse(self._client.coupons)

    @cached_property
    def credit_notes(self) -> credit_notes.AsyncCreditNotesWithStreamingResponse:
        from .resources.credit_notes import AsyncCreditNotesWithStreamingResponse

        return AsyncCreditNotesWithStreamingResponse(self._client.credit_notes)

    @cached_property
    def customers(self) -> customers.AsyncCustomersWithStreamingResponse:
        from .resources.customers import AsyncCustomersWithStreamingResponse

        return AsyncCustomersWithStreamingResponse(self._client.customers)

    @cached_property
    def events(self) -> events.AsyncEventsWithStreamingResponse:
        from .resources.events import AsyncEventsWithStreamingResponse

        return AsyncEventsWithStreamingResponse(self._client.events)

    @cached_property
    def invoice_line_items(self) -> invoice_line_items.AsyncInvoiceLineItemsWithStreamingResponse:
        from .resources.invoice_line_items import AsyncInvoiceLineItemsWithStreamingResponse

        return AsyncInvoiceLineItemsWithStreamingResponse(self._client.invoice_line_items)

    @cached_property
    def invoices(self) -> invoices.AsyncInvoicesWithStreamingResponse:
        from .resources.invoices import AsyncInvoicesWithStreamingResponse

        return AsyncInvoicesWithStreamingResponse(self._client.invoices)

    @cached_property
    def items(self) -> items.AsyncItemsWithStreamingResponse:
        from .resources.items import AsyncItemsWithStreamingResponse

        return AsyncItemsWithStreamingResponse(self._client.items)

    @cached_property
    def metrics(self) -> metrics.AsyncMetricsWithStreamingResponse:
        from .resources.metrics import AsyncMetricsWithStreamingResponse

        return AsyncMetricsWithStreamingResponse(self._client.metrics)

    @cached_property
    def plans(self) -> plans.AsyncPlansWithStreamingResponse:
        from .resources.plans import AsyncPlansWithStreamingResponse

        return AsyncPlansWithStreamingResponse(self._client.plans)

    @cached_property
    def prices(self) -> prices.AsyncPricesWithStreamingResponse:
        from .resources.prices import AsyncPricesWithStreamingResponse

        return AsyncPricesWithStreamingResponse(self._client.prices)

    @cached_property
    def subscriptions(self) -> subscriptions.AsyncSubscriptionsWithStreamingResponse:
        from .resources.subscriptions import AsyncSubscriptionsWithStreamingResponse

        return AsyncSubscriptionsWithStreamingResponse(self._client.subscriptions)

    @cached_property
    def alerts(self) -> alerts.AsyncAlertsWithStreamingResponse:
        from .resources.alerts import AsyncAlertsWithStreamingResponse

        return AsyncAlertsWithStreamingResponse(self._client.alerts)

    @cached_property
    def dimensional_price_groups(self) -> dimensional_price_groups.AsyncDimensionalPriceGroupsWithStreamingResponse:
        from .resources.dimensional_price_groups import AsyncDimensionalPriceGroupsWithStreamingResponse

        return AsyncDimensionalPriceGroupsWithStreamingResponse(self._client.dimensional_price_groups)

    @cached_property
    def subscription_changes(self) -> subscription_changes.AsyncSubscriptionChangesWithStreamingResponse:
        from .resources.subscription_changes import AsyncSubscriptionChangesWithStreamingResponse

        return AsyncSubscriptionChangesWithStreamingResponse(self._client.subscription_changes)


Client = Orb

AsyncClient = AsyncOrb



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel

__all__ = ["ConversionRateTier"]


class ConversionRateTier(BaseModel):
    first_unit: float
    """Exclusive tier starting value"""

    unit_amount: str
    """Amount per unit of overage"""

    last_unit: Optional[float] = None
    """Inclusive tier ending value. If null, this is treated as the last tier"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime
from typing_extensions import Literal

from ...._models import BaseModel
from .affected_block import AffectedBlock
from ...shared.customer_minified import CustomerMinified

__all__ = ["AmendmentLedgerEntry"]


class AmendmentLedgerEntry(BaseModel):
    id: str

    amount: float

    created_at: datetime

    credit_block: AffectedBlock

    currency: str

    customer: CustomerMinified

    description: Optional[str] = None

    ending_balance: float

    entry_status: Literal["committed", "pending"]

    entry_type: Literal["amendment"]

    ledger_sequence_number: int

    metadata: Dict[str, str]
    """User specified key-value pairs for the resource.

    If not present, this defaults to an empty dictionary. Individual keys can be
    removed by setting the value to `null`, and the entire metadata mapping can be
    cleared by setting `metadata` to `null`.
    """

    starting_balance: float



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel
from .custom_expiration import CustomExpiration

__all__ = ["NewAllocationPrice"]


class NewAllocationPrice(BaseModel):
    amount: str
    """An amount of the currency to allocate to the customer at the specified cadence."""

    cadence: Literal["one_time", "monthly", "quarterly", "semi_annual", "annual"]
    """The cadence at which to allocate the amount to the customer."""

    currency: str
    """
    An ISO 4217 currency string or a custom pricing unit identifier in which to bill
    this price.
    """

    custom_expiration: Optional[CustomExpiration] = None
    """The custom expiration for the allocation."""

    expires_at_end_of_cadence: Optional[bool] = None
    """
    Whether the allocated amount should expire at the end of the cadence or roll
    over to the next period. Set to null if using custom_expiration.
    """
